/************************************************************************
 *
 * shuffle.c
 *
 * $Id: shuffle.c,v 1.9 1997/12/09 05:25:25 andrew Exp $
 *
 * SCSI Modifications by Juergen G. Schimmer
 * 1.2 updates by Gareth Jones
 *
 * Copyright 1996, 1997 Andrew Taylor (ataylor@cadvision.com)
 * redistribute freely with copyright intact
 * no warranty, etc.
 *
 ************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/cdrom.h>
#ifdef NEEDS_UCDROM_H
#include <linux/ucdrom.h>
#endif
#include <errno.h>
 
#include "shuffle.h"

#ifdef SCSI_CHANGER
const char *device = "/dev/scd0";
char newdev[128];
#else
const char *device = "/dev/cdrom";
#endif

#if defined(SCSI_CHANGER) || defined(IDE_CHANGER)
const char *def_slots = "0-3";
#else
const char *def_slots = "0";
#endif

int device_fd = -1;
int total_tracks = 0;
int verbose = 0;
int debug = 0;
int shuffle = 1;
int loop = 0;
volatile sig_atomic_t alarm_sounded;
volatile sig_atomic_t skip_ahead;
volatile sig_atomic_t exit_and_stop;
char *pname;			/* program name */

struct {
    int selected;
    int valid;
    int tracks;
} slot[MAX_SLOTS];

struct track_type {
    int slot;
    int slot_track;
    int audio;
    int frames;
} *track;

int main (int argc, char *argv[]) {
    int opt, arg;

    pname = strrchr(argv[0], '/');
    if (pname)
	pname++;
    else
	pname = argv[0];

    while ((opt = getopt (argc, argv, "d:hlsvV")) != -1) {
	switch (opt) {
	  case 'd':
	    device = optarg;
	    break;
	  case 'h':
	    fprintf (stderr, "usage: %s [ -d device ] [ -hlsvV ] [slots...]\n"
		     "\n"
		     "-d device\n\tuse device for cdrom device (default is %s)\n"
		     "-h\thelp\n"
		     "-l\tloop for ever\n"
		     "-s\tnoshuffle\n"
		     "-v\tverbose\n"
		     "-V\tversion\t\n",
		     pname, device);
	    exit(0);
	  case 'l':
	    loop = 1;
	    break;
	  case 's':
	    shuffle = 0;
	    break;
	  case 'v':
	    verbose++;
	    break;
	  case 'V':
	    fprintf (stderr, "shuffle: version %s\n", VERSION);
	    exit(0);
	  default:
	    exit(1);
	}
    }

    arg = optind;

    init_slots();

    if (arg >= argc)
	parse_numeric_arg(def_slots);
    else {
	while (arg < argc) {
	    if (!parse_numeric_arg(argv[arg])) {
		fprintf(stderr, "%s: %s is not a valid argument\n", pname, argv[arg]);
		exit(1);
	    }
	    arg++;
	}
    }

    if (!open_device()) {
	fprintf(stderr, "%s: cannot open device %s: %s\n", pname, device, strerror(errno));
	exit(1);
    }

    get_cd_info();

    do {
       if (shuffle)
	    scramble_tracks();

       play_tracks();
    } while (loop);
	 
    close_device();

    return 0;
}

void init_slots() {
    int i;
    for (i = 0; i < MAX_SLOTS; i++)
	slot[i].selected = 0;
}

/* parse string as decimal number, or as a range seperated by a dash */
int parse_numeric_arg (const char *arg) {
    int i, start, end;

    if (!isdigit(arg[0]))
	return 0;

    i = 0;
    start = arg[i++] - '0';
    while (isdigit(arg[i])) {
	start *= 10;
	start += arg[i++] - '0';
    }

    if (start >= MAX_SLOTS)
	return 0;

    if (arg[i] == '\0') {
	slot[start].selected = 1;
	return 1;
    }

    if (arg[i++] != '-')
	return 0;

    end = arg[i++] - '0';
    while (isdigit(arg[i])) {
	end *= 10;
	end += arg[i++] - '0';
    }
    if (arg[i] != '\0')
	return 0;

    if (end >= MAX_SLOTS)
	return 0;

    for (i = start; i <= end; i++)
	slot[i].selected = 1;

    return 1;
}

int open_device() {
    device_fd = open(device, O_NONBLOCK);
    if (device_fd == -1)
	return 0;
    else
	return 1;
}

void close_device() {
    close(device_fd);
}

void get_cd_info() {
    int i, j;
    struct cdrom_tochdr hdr;
    struct cdrom_tocentry entry;
    int frames, prev_frames;
	
    for (i = 0; i < MAX_SLOTS; i++) {
	if (!slot[i].selected)
	    continue;

	select_cd(i);
	if (ioctl(device_fd, CDROMREADTOCHDR, &hdr) != 0) {
	    slot[i].valid = 0;
	    if (verbose)
		printf ("cd %d: error - skipping\n", i);
	    continue;
	}

	slot[i].valid = 1;
	slot[i].tracks = hdr.cdth_trk1 - hdr.cdth_trk0 + 1;

	if (verbose)
	    printf("cd %d: %d tracks\n", i, slot[i].tracks);

	if (total_tracks == 0)
	    track = malloc (sizeof (struct track_type) * slot[i].tracks);
	else
	    track = realloc (track, sizeof (struct track_type) * (total_tracks + slot[i].tracks));

	/* read starting address of first track */
	read_tocentry(hdr.cdth_trk0, &entry);
	prev_frames = ((entry.cdte_addr.msf.minute * CD_SECS) +
		       entry.cdte_addr.msf.second) * CD_FRAMES +
	               entry.cdte_addr.msf.frame;

	for (j = 0; j < slot[i].tracks; j++) {
	    track[total_tracks + j].slot = i;
	    track[total_tracks + j].slot_track = j + hdr.cdth_trk0;
	    track[total_tracks + j].audio = ((entry.cdte_ctrl & CDROM_DATA_TRACK) == 0);

	    /* read the next toc entry for the start time of the next track */
	    if (j == slot[i].tracks - 1)
		read_tocentry(CDROM_LEADOUT, &entry);
	    else
		read_tocentry(j + hdr.cdth_trk0 + 1, &entry);
	    frames = ((entry.cdte_addr.msf.minute * CD_SECS) +
		      entry.cdte_addr.msf.second) * CD_FRAMES +
		      entry.cdte_addr.msf.frame;

	    track[total_tracks + j].frames = frames - prev_frames;
	    prev_frames = frames;
	}

	total_tracks += slot[i].tracks;
    }
}

/* tell the CD changer to select a new disc */
void select_cd(int n) {
    static int last_slot = -1;
    
    /* if the current disc is being selected, don't bother changing it */
    if (n == last_slot)
	return;
    
#ifdef SCSI_CHANGER
    sprintf(newdev,"/dev/scd%d",n);    
    close(device_fd);
    device_fd = open(newdev,0);
    if (device_fd == -1) {
 	fprintf(stderr, "%s: select disc failed: %s\n", pname, strerror(errno));
 	exit(1);
     }
#elif defined(IDE_CHANGER)
    if (ioctl(device_fd, CDROM_SELECT_DISC, n) == -1) {
	fprintf(stderr, "%s: select disc failed: %s\n", pname, strerror(errno));
	exit(1);
    }
#endif

    last_slot = n;
}

void read_tocentry(int trackno, struct cdrom_tocentry *entry) {
    entry->cdte_track = trackno;
    entry->cdte_format = CDROM_MSF;
    
    if (ioctl(device_fd, CDROMREADTOCENTRY, entry)) {
	fprintf(stderr, "%s: unable to read table of contents: %s\n",
		pname, strerror(errno));
	exit(1);
    }
}

/* shuffle the tracks list */
void scramble_tracks() {
    int i, j;
    int r;
    struct track_type tmp;

    srand((unsigned int) time(NULL));

    for (i = 0; i < SHUFFLES; i++) {
	for (j = 0; j < total_tracks; j++) {
	    r = (int) (10.0*rand() / (RAND_MAX+1.0));
	    tmp = track[j];
	    track[j] = track[r];
	    track[r] = tmp;
	}
    }
}

/* play all the tracks in the global `tracks' array */
void play_tracks() {
    int i;
    struct cdrom_ti ti;

    signal(SIGALRM, sig_handler);
    signal(SIGUSR1, sig_handler);
    signal(SIGTERM, sig_handler);
    signal(SIGINT, sig_handler);

    for (i = 0; i < total_tracks; i++) {

	if (!track[i].audio) {
	    if (verbose)
		printf("cd %d track %d: skipping data track\n",
		       track[i].slot, track[i].slot_track);
	    continue;
	}

	select_cd(track[i].slot);

	if (verbose)
	    printf("cd %d track %d\n", track[i].slot, track[i].slot_track);

	ti.cdti_trk0 = track[i].slot_track;
	ti.cdti_ind0 = 1;
#ifdef SCSI_CHANGER
	ti.cdti_trk1 = (track[i].slot_track +1);
#else
	ti.cdti_trk1 = track[i].slot_track;
#endif
	ti.cdti_ind1 = 1;
	
	if (ioctl(device_fd, CDROMSTART, 0) == -1)
	    fprintf(stderr, "%s: error on start: %s\n", pname, strerror(errno));

	if (ioctl(device_fd, CDROMPLAYTRKIND, &ti) == -1)
	    fprintf(stderr, "%s: error on play: %s\n", pname, strerror(errno));

	/* wait for track to play */
	if (wait_track(i))
	    break;
    }

    ioctl(device_fd, CDROMSTOP);
}

/* wait for the given track to finish, then return; return non-zero if
   program should stop immediately */
int wait_track (int trackno) {
    struct itimerval timer;
    sigset_t mask, oldmask;

    alarm_sounded = 0;
    skip_ahead = 0;
    exit_and_stop = 0;

    /* set an alarm when the track should be done */
    timer.it_interval.tv_sec = 0;
    timer.it_interval.tv_usec = 0;
    timer.it_value.tv_sec = track[trackno].frames / CD_FRAMES;
    timer.it_value.tv_usec = (track[trackno].frames % CD_FRAMES) * (1000000 / CD_FRAMES);
    setitimer(ITIMER_REAL, &timer, NULL);

    /* wait for ALARM, TERM, INT, USR1 signals */
    sigemptyset (&mask);
    sigaddset (&mask, SIGALRM);
    sigaddset (&mask, SIGTERM);
    sigaddset (&mask, SIGINT);
    sigaddset (&mask, SIGUSR1);

    sigprocmask(SIG_BLOCK, &mask, &oldmask);
    while (!alarm_sounded && !skip_ahead && !exit_and_stop)
	sigsuspend(&oldmask);
    sigprocmask(SIG_UNBLOCK, &mask, NULL);

    /* if the signal wasn't ALARM, cancel the alarm */
    if (!alarm_sounded) {
	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = 0;
	timer.it_value.tv_sec = 0;
	timer.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &timer, NULL);
    }

    return exit_and_stop;
}

/* this handler is used for all (trapped) signals */
void sig_handler(int sig) {
    signal(sig, sig_handler);
    switch (sig) {
      case SIGTERM:
      case SIGINT:
	exit_and_stop = 1;
	loop = 0;
	break;
      case SIGALRM:
	alarm_sounded = 1;
	break;
      case SIGUSR1:
	skip_ahead = 1;
	break;
      case SIGUSR2:
	break;
      default:
	abort();
    }
}
